%Kalman filter algorithm tracking a beam and estimating excitation
close all;clc;

%measurements
alt_d = 97; %altitude of displacement sensor
alt_a = 64; %altitude of accelerometer
mea_dis = xlsread('position_X_97m_100sec.xlsx');
mea_acc = xlsread('acceleration_X_64m_100sec.xlsx');
mea = [mea_dis(:,2)' ; mea_acc(:,2)'];

t = linspace(0,100,length(mea));
dt = (t(end) - t(1))/(length(t)-1); %Time step%

%System parameters
%Geometry 
geometry.L = 107.6; % beam length (m)
geometry.E = 2.1e11; % Young Modulus (Pa)
geometry.nu = 0.3; % Poisson ratio
geometry.rho = 8500; % density (kg/m^3)
geometry.t = 0.06 ; %thikness of the tube (m)

% case of a tube
D = 6 ; % beam diameter (m)
d = D - geometry.t ; % Inner diameter of the tube (m)
% beam is symmetrical around axes y and z
Iy = pi.*(D.^4 - d.^4)./64;
Ix = pi.*(D.^4 - d.^4)./64;
geometry.I = Iy; % affectation of quadratic moment

%Mass at the tip of the beam
geometry.Mtip = 400*10^(3) ; %(kg) 

ndisc = 108 ; % number of discretisation points for the beam
geometry.y = linspace(0,geometry.L ,ndisc);
V = ((pi.*(D/2).^2)-(pi.*(d/2).^2))*geometry.L;
geometry.m = (geometry.rho.*V)./geometry.L;


% Number of modes
Nmodes =3; % number of mode wanted
[phi,wn] = eigenModes(geometry,Nmodes);
%Eigen frequencies
Omega = diag(wn);
%Damping ratio
Gamma = 2/100*diag([1.89 1.38 0.89]).*Omega;

%Excitation force location
np = 1; %Number of forces applying on the structure
Sp = zeros(ndisc, np);
Sp(ndisc,1) = 1; %1 force located on the tip of the beam
%Measurement transmission matrices
nd = 2; %number of measurement (disp, vel and acc)
Sdd = zeros(nd,ndisc);
Sdd(1,alt_d) = 1 ; %Displacement measurement taken at 20 m from the bottom
Sdv = zeros(nd,ndisc); %No velocity measurement
Sda = zeros(nd,ndisc);
Sda(2,alt_a) = 1 ; %Acceleration measurement taken at 20 m from the tip

%Modeling errors
%Geometry 
geometry_err.L = 1*geometry.L; % beam length (m)
geometry_err.E = (1+0.05)*geometry.E; % Young Modulus (Pa)
geometry_err.nu = 1*geometry.nu; % Poisson ratio
geometry_err.rho = (1-0.01)*geometry.rho; % density (kg/m^3)
geometry_err.t = 1*geometry.t; % thikness (m)


% case of a tube
D_err = 6; % beam diameter (m)
d_err = D_err - geometry_err.t ; % Inner diameter of the tube (m)
% beam is symmetrical around axes y and z
Iy_err =  pi.*(D_err.^4 - d_err.^4)./64;
Ix_err =  pi.*(D_err.^4 - d_err.^4)./64;
geometry_err.I = Iy_err; % affectation of quadratic moment

geometry_err.y = linspace(0,geometry_err.L ,ndisc);
V_err = ((pi.*(D_err/2).^2)-(pi.*(d_err/2).^2))*geometry_err.L;
geometry_err.m = geometry_err.rho.*V_err./geometry_err.L;

% Number of modes
Nmodes =3; % number of mode wanted

[phi_err,wn_err] = eigenModes(geometry_err,Nmodes);

%Eigen frequencies
Omega_err = diag(wn_err);
%Damping ratio
Gamma_err = (1+0.1)*Gamma;

%State space model for model error 
ns = 2*Nmodes;
A_err = exp([zeros(Nmodes) eye(Nmodes); -Omega_err.^2 -Gamma_err]*dt);
B_err = (A_err - eye(ns))/[zeros(Nmodes) eye(Nmodes); -Omega_err.^2 -Gamma_err]*[zeros(Nmodes,np); phi_err*Sp];
G_err = [Sdd*phi_err' - Sda*phi_err'*Omega_err^2 , Sdv*phi_err' - Sda*phi_err'*Gamma_err];
J_err = (Sda*(phi_err'*phi_err)*Sp); 

%State space model
A = exp([zeros(Nmodes) eye(Nmodes); -Omega.^2 -Gamma]*dt);
B = (A - eye(ns))/[zeros(Nmodes) eye(Nmodes); -Omega.^2 -Gamma]*[zeros(Nmodes,np); phi*Sp];
G = [Sdd*phi' - Sda*phi'*Omega^2 , Sdv*phi' - Sda*phi'*Gamma];
J = (Sda*(phi'*phi)*Sp); 

delta_A = abs(A_err - A);
delta_B = abs(B_err - B);
delta_G = abs(G_err - G);
delta_J = abs(J_err - J);

[x_est,f_est] = Joint_input_state_beam_func(mea,alt_d,alt_a);

Cp = var(f_est)*eye(np,np);
Cx =diag(var(x_est'));
Cxp = zeros(ns,np);
for i=1:ns
    Cxp(i) = [1 0]*cov(x_est(i,:),f_est)*[0;1];
end
Cpx = Cxp';

%Stochastic forces
nps = ndisc; %Number of stochastic forces
ps = ones(nps,1); %Stochastic forces distributed all along the beam
ps(1:19) = 1000*ones(19,1);
ps(20) = 5000;
ps(21:nps) = 100*ones(nps-20,1);
Cps = diag(ps).^2; %Stochastic forces of 10 N standart deviation
Bs = zeros(ns,nps);
Js = zeros(nd,nps);
for i=1:nps
    Sps = zeros(nps,1);
    Sps(i) = ps(i);
    Bs(:,i)=(A - eye(ns))/[zeros(Nmodes) eye(Nmodes); -Omega.^2 -Gamma]*[zeros(Nmodes,np); phi*Sps];
    Js(:,i)=(Sda*(phi'*phi)*Sps);
end

%error covariance matrices
Rm = zeros(nd);
Rm(1,1) = 10^(-1); %measurement error on first measure
Rm(2,2) = 10^(-1); %measurement error on second measure
%Rm(3,3) = ... %measurement error on third measure etc.
Q = Bs*Cps*Bs' + delta_A*Cx*delta_A' + delta_B*Cp*delta_B' + delta_A*Cxp*delta_B' + delta_B*Cpx*delta_A';
S = Bs*Cps*Js' + delta_A*Cx*delta_G' + delta_B*Cp*delta_J' + delta_A*Cxp*delta_J' + delta_B*Cpx*delta_G';
R = Js*Cps*Js' + Rm + delta_G*Cx*delta_G' + delta_J*Cp*delta_J' + delta_G*Cxp*delta_J' + delta_J*Cpx*delta_G' ;

%Output vector
nde = 1; %Number of output to be estimated
Sdde = zeros(nde,ndisc);
Sdde(1,100) = 1 ; %Displacement measurement taken at the tip of the beam
Sdve = zeros(nde,ndisc); %No velocity output
Sdae = zeros(nde,ndisc); %No acceleration output
Ge = [Sdde*phi' - Sdae*phi'*Omega^2 , Sdve*phi' - Sdae*phi'*Gamma];
Je = (Sdae*(phi'*phi)*Sp);
Jes = zeros(nde,nps); %Relating output vector to stochastic forces

%output covariance matrices
Re = Jes*Cps*Jes';
Rc = Jes*Cps*Js';

%Pre-definition of variables
x = zeros(ns,length(mea));
x_e = zeros(ns,length(mea));
Px = zeros(ns,ns);
Pp = zeros(np,np); %size of J matrix
f = zeros(np,length(mea));
de = zeros(nde,length(mea));

%Initialization
y0 = phi*mea(1,1)*ones(ndisc,1); %Initial position (modal coordinates) for each element 
v0 = phi*mea(2,1)*ones(ndisc,1); %Initial velocity (modal coordinates) for each element
Px0 = zeros(size(Px));
x(:,1) = [y0; v0]; 
Px = Px0;
Px_e = Px;

%Looping process 
tic
for i=2:length(mea)
    %Input estimation 
    R_k = G*Px_e*G'+R;
    M_k = (J'/R_k*J)\(J'/R_k);
    f(i) = M_k*(mea(:,i)-G*x(:,i-1));
    Pp = eye(size(Pp))/(J'/R_k*J);
    %Measurement update
    K = Px_e*G'/R_k;
    x_e(:,i) = x(:,i-1)+K*(mea(:,i)-G*x(:,i-1)-J*f(i)); 
    Px = Px_e-K*(R_k-J*Pp*J')*K';
    Pxp = -K*J*Pp;
    %Time update 
    x(:,i) = A*x_e(:,i)+B*f(i);
    I=eye(size(J*M_k));
    N_k = A*K*(I-J*M_k)+B*M_k;
    Px_e = [A B]*[Px Pxp;Pxp' Pp]*[A';B']+Q+N_k*S'+S*N_k';
   
    %Output vector
    de(i) = Ge*x_e(:,i)+Je*f(i);
    Z = Ge*K*(I-J*M_k)+Je*M_k;
    Pde = Ge*Px*Ge'+Je*Pp*Je'+Ge*Pxp*Je'+Je*Pxp'*Ge'+Re-Z*Rc'-Rc*Z';
end
toc